from .errors import EmptyArray
import numpy as np


class ArrayMapper:
    """Map an array of values to integer indices ascending from 0, and store this mapping for application to other arrays of floats.

    This code improves on previous approaches (see e.g. https://github.com/brightway-lca/brightway2-calc/blob/114c272a0fde9301ff5e250e2773c01c28d7fd99/bw2calc/indexing.py) by:

    * Not using ``np.searchsorted``, which appears to be always slower than ``index_with_arrays`` (see ``dev`` folder).
    * Keeping the indexing arrays in memory, avoiding their creation each time they are applied to a new matrix.

    .. code-block:: python

        In [1]: from matrix_utils import ArrayMapper
           ...: import numpy as np

        In [2]: am = ArrayMapper(array=np.array([0, 4, 6]))

        In [3]: am.map_array(np.array([6, 6, 3, 0]))
        Out[3]: array([ 2,  2, -1,  0])

    """

    def __init__(
        self, *, array: np.ndarray, sparse_cutoff: float = 0.1, empty_ok: bool = False
    ):
        self._check_input_array(array)

        # Even if already unique, this only adds ~2ms for 100.000 elements
        self.array = np.unique(array)

        # TODO
        # Sparse matrices could be used if the number of values present is much less
        # than the number of possible values, given the (min, max) range.
        # The default code will generate a complete mapping for the (min, max)
        # interval, which can use too much memory in certain cases.
        # self.use_sparse = len(self.keys) / self.keys.max() <= sparse_cutoff:

        if array.shape == (0,):
            if empty_ok:
                self.max_value = 0
            else:
                raise EmptyArray("Empty array can't be used to map values")
        else:
            self.max_value = int(self.array.max())
        self.index_array = np.zeros(self.max_value + 1) - 1
        self.index_array[self.array] = np.arange(len(self.array))

    def __len__(self):
        return self.array.shape[0]

    def _check_input_array(self, array: np.ndarray) -> None:
        if len(array.shape) != 1:
            raise ValueError("array must be 1-d")
        if array.shape[0] and array.min() < 0:
            raise ValueError("Array index values must be positive")

    def map_array(self, array: np.ndarray) -> np.ndarray:
        self._check_input_array(array)

        if array.shape == (0,):
            # Empty array
            return array.copy()

        result = np.zeros_like(array) - 1
        mask = array <= self.max_value
        result[mask] = self.index_array[array[mask]]
        return result

    def to_dict(self) -> dict:
        """Turn the mapping arrays into a Python dict. This is only useful for
        human examination, the normal implementation uses Numpy functions on the
        arrays directly."""
        return {int(x): int(y) for x, y in zip(self.array, self.map_array(self.array))}

    def reverse_dict(self) -> dict:
        return {y: x for x, y in self.to_dict().items()}



from .array_mapper import ArrayMapper
from .indexers import RandomIndexer, SequentialIndexer, CombinatorialIndexer
from .resource_group import ResourceGroup
from .utils import filter_groups_for_packages, safe_concatenate_indices
from bw_processing import Datapackage
from scipy import sparse
from typing import Union, Sequence, Any, Callable
import numpy as np


class MappedMatrix:
    """A scipy sparse matrix handler which takes in ``bw_processing`` data packages. Row and column ids are mapped to matrix indices, and a matrix is constructed.

    `indexer_override` allows for custom indexer behaviour. Indexers should follow a simple API: they must support `.__next__()`, and have the attribute `.index`, which returns an integer.

    `custom_filter` allows you to remove some data based on their indices. It is applied to all resource groups. If you need more fine-grained control, process the matrix after construction/iteration. `custom_filter` should take the indices array as an input, and return a Numpy boolean array with the same length as the indices array.

    Args:

        * packages: A list of Ddatapackage objects.
        * matrix: The string identifying the matrix to be built.
        * use_vectors: Flag to use vector data from datapackages
        * use_arrays: Flag to use array data from datapackages
        * use_distributions: Flag to use `stats_arrays` distribution data from datapackages
        * row_mapper: Optional instance of `ArrayMapper`. Used when matrices must align.
        * col_mapper: Optional instance of `ArrayMapper`. Used when matrices must align.
        * seed_override: Optional integer. Overrides the RNG seed given in the datapackage, if any.
        * indexer_override: Parameter for custom indexers. See above.
        * diagonal: If True, only use the `row` indices to build a diagonal matrix.
        * custom_filter: Callable for function to filter data based on `indices` values. See above.
        * empty_ok: If False, raise `AllArraysEmpty` if the matrix would be empty

    """

    def __init__(
        self,
        *,
        packages: Sequence[Datapackage],
        matrix: str,
        use_vectors: bool = True,
        use_arrays: bool = True,
        use_distributions: bool = False,
        row_mapper: Union[ArrayMapper, None] = None,
        col_mapper: Union[ArrayMapper, None] = None,
        seed_override: Union[int, None] = None,
        indexer_override: Any = None,
        diagonal: bool = False,
        custom_filter: Union[Callable, None] = None,
        empty_ok: bool = False,
    ):
        self.seed_override = seed_override
        self.diagonal = diagonal
        self.matrix_label = matrix
        self.packages = {
            package: [
                ResourceGroup(
                    package=filtered_package,
                    group_label=group_label,
                    use_distributions=use_distributions,
                    seed_override=seed_override,
                    custom_filter=custom_filter,
                )
                for group_label, filtered_package in lst
            ]
            for package, lst in filter_groups_for_packages(
                packages, matrix, use_vectors, use_arrays, use_distributions
            ).items()
        }
        self.groups = tuple([obj for lst in self.packages.values() for obj in lst])
        self.add_indexers(indexer_override, seed_override)

        self.row_mapper = row_mapper or ArrayMapper(
            array=safe_concatenate_indices(
                [obj.unique_row_indices() for obj in self.groups], empty_ok
            ),
            empty_ok=empty_ok,
        )
        if not diagonal:
            self.col_mapper = col_mapper or ArrayMapper(
                array=safe_concatenate_indices(
                    [obj.unique_col_indices() for obj in self.groups], empty_ok
                ),
                empty_ok=empty_ok,
            )

        self.add_mappers(
            axis=0,
            mapper=self.row_mapper,
        )

        if not diagonal:
            self.add_mappers(axis=1, mapper=self.col_mapper)
        self.map_indices()

        row_indices = safe_concatenate_indices([obj.row for obj in self.groups], empty_ok)
        col_indices = safe_concatenate_indices([obj.col for obj in self.groups], empty_ok)

        if diagonal:
            x = int(self.row_mapper.index_array.max() + 1)
            dimensions = (x, x)
        else:
            dimensions = (
                int(self.row_mapper.index_array.max() + 1),
                int(self.col_mapper.index_array.max() + 1),
            )

        self.matrix = sparse.coo_matrix(
            (
                np.zeros(len(row_indices)),
                (row_indices, col_indices),
            ),
            dimensions,
            dtype=np.float64,
        ).tocsr()

        self.rebuild_matrix()

    def add_mappers(self, axis: int, mapper: ArrayMapper):
        for group in self.groups:
            group.add_mapper(axis=axis, mapper=mapper)

    def map_indices(self):
        for group in self.groups:
            group.map_indices(diagonal=self.diagonal)

    def iterate_indexers(self):
        for obj in self.packages:
            if hasattr(obj, "indexer"):
                next(obj.indexer)

    def rebuild_matrix(self):
        self.matrix.data *= 0
        for group in self.groups:
            row, col, data = group.calculate()
            if group.package.metadata["sum_inter_duplicates"]:
                self.matrix[row, col] += data
            else:
                self.matrix[row, col] = data

    def __next__(self):
        self.iterate_indexers()
        self.rebuild_matrix()

    def add_indexers(self, indexer_override: Any, seed_override: Union[int, None]):
        """Add indexers"""
        for package, resources in self.packages.items():
            if hasattr(package, "indexer"):
                for obj in resources:
                    obj.add_indexer(indexer=package.indexer)
            elif indexer_override is not None:
                package.indexer = indexer_override
                for obj in resources:
                    obj.add_indexer(indexer=package.indexer)
            elif package.metadata["combinatorial"]:
                package.indexer = CombinatorialIndexer(
                    [obj.ncols for obj in resources if obj.ncols]
                )
                for i, obj in enumerate(resources):
                    obj.add_combinatorial_indexer(indexer=package.indexer, offset=i)
            elif package.metadata["sequential"] and seed_override is None:
                package.indexer = SequentialIndexer()
                for obj in resources:
                    obj.add_indexer(indexer=package.indexer)
            else:
                package.indexer = RandomIndexer(
                    seed=seed_override or package.metadata["seed"]
                )
                for obj in resources:
                    obj.add_indexer(indexer=package.indexer)



class MatrixUtilsError(Exception):
    """Base class for bw2calc errors"""

    pass


class AllArraysEmpty(MatrixUtilsError):
    """Can't load the numpy arrays if all of them are empty"""

    pass


class NoArrays(MatrixUtilsError):
    """No arrays for given matrix"""

    pass


class EmptyArray(MatrixUtilsError):
    """Empty array can't be used"""

    pass


class IncompatibleDataTypes(MatrixUtilsError):
    """Data types present in resource group can't be both present"""

    pass



from .errors import AllArraysEmpty
import numpy as np


def filter_groups_for_packages(
    packages, matrix_label, use_vectors, use_arrays, use_distributions
) -> dict:
    return {
        package: [
            (group_label, filtered_package)
            for group_label, filtered_package in package.groups.items()
            if has_relevant_data(
                group_label, package, use_vectors, use_arrays, use_distributions
            )
        ]
        for package in [
            obj.filter_by_attribute("matrix", matrix_label) for obj in packages
        ]
    }


def has_relevant_data(
    group_label, package, use_vectors, use_arrays, use_distributions
) -> bool:
    return any(
        res
        for res in package.resources
        if res["group"] == group_label
        and (
            (res["kind"] == "data" and res["category"] == "vector" and use_vectors)
            or (
                res["kind"] == "distributions"
                and res["category"] == "vector"
                and use_distributions
            )
            # Use vectors under Monte Carlo as fallback. Warning: Could be changed in future!
            or (
                res["kind"] == "data"
                and res["category"] == "vector"
                and use_distributions
            )
            or (res["kind"] == "data" and res["category"] == "array" and use_arrays)
        )
    )


def safe_concatenate_indices(
    arrays: [np.ndarray], empty_ok: bool = False
) -> np.ndarray:
    try:
        return np.hstack(arrays)
    except ValueError:
        if empty_ok:
            return np.array([], dtype=int)
        else:
            raise AllArraysEmpty



__all__ = (
    "__version__",
    "ResourceGroup",
    "ArrayMapper",
    "MappedMatrix",
    "RandomIndexer",
    "SequentialIndexer",
    "CombinatorialIndexer",
    "Proxy",
)

from .version import version as __version__
from .array_mapper import ArrayMapper
from .mapped_matrix import MappedMatrix
from .resource_group import ResourceGroup
from .indexers import RandomIndexer, SequentialIndexer, CombinatorialIndexer, Proxy



from itertools import product
from numpy.random import Generator, PCG64
from typing import Union, List


# Max signed 32 bit integer, compatible with Windows
MAX_SIGNED_32BIT_INT = 2147483647


class Indexer:
    """Base class for indexers"""

    pass


class RandomIndexer(Generator, Indexer):
    """A (potentially) seeded integer RNG that remembers the generated index.

    Returns indices for a sample array.

    max_value: Number of columns in the array for which a column index is returned.
    seed: Seed for RNG. Optional."""

    def __init__(self, seed: Union[int, None] = None):
        super().__init__(PCG64(seed))
        next(self)

    def __next__(self):
        self.index = self.integers(0, MAX_SIGNED_32BIT_INT)
        return self.index


class SequentialIndexer(Indexer):
    def __init__(self):
        self.index = 0

    def __next__(self):
        self.index += 1
        return self.index


class CombinatorialIndexer(Indexer):
    def __init__(self, max_values: List[int]):
        self.iterator = product(*[range(x) for x in max_values])
        next(self)

    def __next__(self):
        self.index = next(self.iterator)
        return self.index


class Proxy:
    """Simple class that proxies access to a combinatorial indexer"""

    def __init__(self, indexer: CombinatorialIndexer, offset: int):
        self.indexer = indexer
        self.offset = offset

    @property
    def index(self):
        return self.indexer.index[self.offset]



from .aggregation import aggregate_with_sparse
from .array_mapper import ArrayMapper
from .errors import IncompatibleDataTypes
from .indexers import Indexer, Proxy
from bw_processing import DatapackageBase
from stats_arrays import MCRandomNumberGenerator
from typing import Any, Union, Callable
import numpy as np


class FakeRNG:
    def __init__(self, array):
        self.array = array

    def __next__(self):
        return self.array


class ResourceGroup:
    """A class that handles a resource group - a collection of files in data package which define one matrix. A resource group can contain the following:

    * data array or interface (required)
    * indices structured array (required)
    * flip array (optional)
    * csv metadata (optional)
    * json metadata (optional)

    After instantiation, the ``MappedMatrix`` class will add an indexer (an instance of ``matrix_utils.indexers.Indexer``), using either ``add_indexer`` or ``add_combinatorial_indexer``. It will also add one or two array mappers, using ``add_mapper``. Only one mapper is needed if the matrix is diagonal.

    One easy source of confusion is the difference between ``.row_original`` and ``.row``. Both of these are one-dimensional vectors which have matrix row indices (i.e. they have already been mapped). ``.row_original`` are the values as given in the data package, whereas ``.row`` are the values actually used in matrix data insertion (and the same for column indices).There are two possible modifications that can be applied from ``.row_original`` to ``.row``; first, we can have internal aggregation, which will shrink the size of the row or column indices, as duplicate elements are eliminated. Second, if the array mapper was already instantiated, we might need to delete some elements from the row or column indices, as these values are not used in this calculation. For example, in life cycle assessment, LCIA method implementations often contain characterization factors for flows not present in the biosphere (as they were not used by an of the activities). In this case, we would need to eliminate these factors, as our characterization matrix must match exactly to the biosphere matrix already built.

    Here is an example for row indices:

    .. code-block:: python

        row_input_indices = [0, 17, 99, 42, 17]
        row_original = [0, 1, -1, 2, 1]
        after_aggregation = [0, 1, -1, 2]  # -1 is missing data point
        after_masking = [0, 1, 2]
        row = [0, 1, 2]

    Any data coming into this class, with through instantiation or via method calls such as ``.calculate``, should follow the length and order of ``.row_original``.

    The current data, as entered into the matrix, is given by ``.current_data``.

    """

    def __init__(
        self,
        *,
        package: DatapackageBase,
        group_label: str,
        use_distributions: bool = False,
        seed_override: Union[int, None] = None,
        custom_filter: Union[Callable, None] = None,
    ):
        self.label = group_label
        self.package = package
        self.use_distributions = use_distributions
        self.custom_filter = custom_filter
        self.vector = self.is_vector()

        if custom_filter is not None:
            self.filter_mask = custom_filter(self.raw_indices)
        else:
            self.filter_mask = None

        if self.use_distributions and not self.vector:
            raise IncompatibleDataTypes
        elif self.use_distributions:
            seed = seed_override or self.package.metadata.get("seed")
            if self.has_distributions:
                self.rng = MCRandomNumberGenerator(params=self.data, seed=seed)
            else:
                self.rng = FakeRNG(self.data)

        self.aggregate = self.package.metadata["sum_intra_duplicates"]
        self.empty = self.indices.shape == (0,)

    @property
    def has_distributions(self):
        try:
            self.get_resource_by_suffix("distributions")
            return True
        except KeyError:
            return False

    @property
    def data(self):
        if self.use_distributions and self.has_distributions:
            return self.get_resource_by_suffix("distributions")
        else:
            return self.get_resource_by_suffix("data")

    @property
    def raw_flip(self):
        """The source data for the flip array."""
        return self.get_resource_by_suffix("flip")

    @property
    def flip(self):
        """The flip array, with the custom filter mask applied if necessary."""
        if self.filter_mask is None:
            return self.raw_flip
        else:
            return self.raw_flip[self.filter_mask]

    @property
    def raw_indices(self):
        """The source data for the indices array."""
        return self.get_resource_by_suffix("indices")

    @property
    def indices(self):
        """The indices array, with the custom filter mask applied if necessary."""
        if self.filter_mask is None:
            return self.raw_indices
        else:
            return self.raw_indices[self.filter_mask]

    def get_resource_by_suffix(self, suffix: str) -> Any:
        return self.package.get_resource(self.label + "." + suffix)[0]

    def is_vector(self) -> bool:
        """Determine if this is a vector or array resource"""
        metadata = self.package.get_resource(self.label + ".data")[1]
        return metadata.get("category") == "vector"

    @property
    def ncols(self):
        if self.vector:
            return None
        else:
            return self.data.shape[1]

    def add_mapper(self, axis: int, mapper: ArrayMapper):
        if axis == 0:
            self.row_mapper = mapper
        else:
            self.col_mapper = mapper

    def build_mask(self, row, col):
        """Build boolean array mask where ``False`` means that a data element is not present, and should be ignored. See discussion above."""
        mask = (row != -1) * (col != -1)
        if (~mask).sum():
            return mask
        else:
            return None

    def map_indices(self, *, diagonal=False):
        if self.empty:
            self.row = np.array([])
            self.col = np.array([])
            return

        self.row_original = self.row_mapper.map_array(self.indices["row"])
        if diagonal:
            self.col_original = self.row_original
        else:
            self.col_original = self.col_mapper.map_array(self.indices["col"])

        self.mask = self.build_mask(self.row_original, self.col_original)

        if self.aggregate:
            self.count = max(self.row_original.max(), self.col_original.max()) + 1
            if self.mask is not None:
                self.row, self.col, _ = aggregate_with_sparse(
                    self.row_original[self.mask],
                    self.col_original[self.mask],
                    np.zeros(self.mask.sum()),
                    self.count,
                )
            else:
                self.row, self.col, _ = aggregate_with_sparse(
                    self.row_original,
                    self.col_original,
                    np.zeros(self.row_original.shape),
                    self.count,
                )

        else:
            if self.mask is not None:
                self.row = self.row_original[self.mask]
                self.col = self.col_original[self.mask]
            else:
                self.row = self.row_original
                self.col = self.col_original

    def unique_row_indices(self):
        """Return array of unique indices that respect aggregation policy"""
        return np.unique(self.indices["row"])

    def unique_col_indices(self):
        """Return array of unique indices that respect aggregation policy"""
        return np.unique(self.indices["col"])

    def add_indexer(self, indexer: Indexer):
        self.indexer = indexer

    def add_combinatorial_indexer(self, indexer: Indexer, offset: int):
        self.indexer = Proxy(indexer, offset)

    def calculate(self, vector: np.ndarray = None):
        """Generate row and column indices and a data vector. If ``.data`` is an iterator, draw the next value. If ``.data`` is an array, use the column given by ``.indexer``.

        ``vector`` is an optional input that overrides the data. It must be in the same order and have the same length as the data package indices (before possible aggregation and masking); see discussion above.

        """
        if self.empty:
            self.current_data = np.array([])
            return self.row, self.col, self.current_data

        if vector is not None:
            data = vector
        elif self.vector:
            if self.use_distributions:
                data = next(self.rng)
            else:
                try:
                    data = next(self.data)
                except TypeError:
                    data = self.data
        else:
            data = self.data[:, self.indexer.index % self.ncols]

        # Copy to avoid modifying original data
        data = data.copy()

        if self.filter_mask is not None:
            data = data[self.filter_mask]

        try:
            data[self.flip] *= -1
        except KeyError:
            # No flip array
            pass

        # Second copy because we want to store the data before aggregation
        self.current_data = data.copy()

        if self.aggregate:
            if self.mask is not None:
                return aggregate_with_sparse(
                    self.row_original[self.mask],
                    self.col_original[self.mask],
                    data[self.mask],
                    self.count,
                )
            else:
                return aggregate_with_sparse(
                    self.row_original, self.col_original, data, self.count
                )
        else:
            return self.row, self.col, data[self.mask]



from scipy import sparse
import numpy as np


def aggregate_with_sparse(
    rows: np.ndarray, cols: np.ndarray, data: np.ndarray, count: int
):
    # About three times faster than pandas groupby for our example use case (100.000 data points, 500 possible row and column indices)
    # See dev/speed_test.py
    # and https://jakevdp.github.io/PythonDataScienceHandbook/03.08-aggregation-and-grouping.html for other use cases

    # Note that the conversion to CSR is necessary for aggregation
    matrix = sparse.coo_matrix((data, (rows, cols)), (count, count)).tocsr().tocoo()
    return matrix.row, matrix.col, matrix.data



from setuptools import setup

v_temp = {}
with open("matrix_utils/version.py") as fp:
    exec(fp.read(), v_temp)
version = ".".join((str(x) for x in v_temp["version"]))


setup(
    name="matrix_utils",
    version=version,
    packages=["matrix_utils"],
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    license="BSD 3-clause",
    install_requires=[
        "numpy",
        "scipy",
        "pandas",
        "bw_processing",
        "stats_arrays",
    ],
    url="https://github.com/brightway-lca/matrix_utils",
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    description="Tools to create matrices from data packages",
    classifiers=[
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
)
